/*!
 * © 2022 David Vines
 *
 * domainOfTheAncients is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * any later version.
 *
 * domainOfTheAncients is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with domainOfTheAncients. If not, see https://www.gnu.org/licenses/gpl-2.0.html.
 */
/* globals module, exports, ShipDesign, ShipDesignType, Block, ShipOrder, ROT */
/* exported Ship */
class Ship {
	constructor(name,design,startLocation) {
		this._name = name;
		this._design = design;
		this._location = startLocation;
		this._order = new ShipOrder();
		this._report = "Nothing to report";
		this._blocks = null; // No cached field yet
		this._powerNeedsCalculating = true;
		this._tractoredPoints = 0;
		this._shields = 0;
		this._reflectingShields = 0;
		this._reflectedDamageThisPhase = 0;
		this._reflectedDamageNextPhase = 0;
		this._retreatingTo = null;
		this._maxMarines = this._design.getMarinesOnBoard();
		this._marinesOnBoard = this._maxMarines;
	}
	get name() { return this._name; }
	get design() { return this._design; }
	get classname() { return this._design.name; }
	get location() { return this._location; }
	set location(location) { this._location = location; }
	get report() { return this._report;  }
	set report(report) { this._report = report; }
	resetReport() { this._report = ""; }
	hasAReport() { return this._report !== ""; }
	get order() { return this._order; }
	set order(order) { this._order = order; }
	get lastHex() { return this._lastHex; }
	set lastHex(hex) { this._lastHex = hex; }
	get retreatingTo() { return this._retreating; }
	set retreatingTo(hex) { this._retreatingTo = hex; }
	getJumpRange() {
		this._powerSystems();
		let jumpPoints = 0;
		for(let block of this._blocks) {
			if (block.key == "_ja" && block.operational) jumpPoints += 10;
			if (block.key == "_jb" && block.operational) jumpPoints += 20;
			if (block.key == "_jc" && block.operational) jumpPoints += 30;
			if (block.key == "_jd" && block.operational) jumpPoints += 40;
			if (block.key == "_je" && block.operational) jumpPoints += 50;
			if (block.key == "_jf" && block.operational) jumpPoints += 75;
			if (block.key == "_jg" && block.operational) jumpPoints += 100;
			if (block.key == "_jh" && block.operational) jumpPoints += 200;
		}
		return Math.floor(jumpPoints/this._design.totalSize());
	}
	getCombatSpeed() {
		this._powerSystems();
		let manPoints = -this._tractoredPoints;
		for(let block of this._blocks) {
			if (block.key == "_ja" && block.operational) manPoints += 10;
			if (block.key == "_jb" && block.operational) manPoints += 20;
			if (block.key == "_jc" && block.operational) manPoints += 30;
			if (block.key == "_jd" && block.operational) manPoints += 40;
			if (block.key == "_je" && block.operational) manPoints += 50;
			if (block.key == "_jf" && block.operational) manPoints += 75;
			if (block.key == "_jg" && block.operational) manPoints += 100;
			if (block.key == "_jh" && block.operational) manPoints += 200;
		}
		if (manPoints < 0) { manPoints = 0; }
		return Math.floor(manPoints/this._design.totalSize()) + this.getBestComputer();
	}
	getBestComputer() { // Note that a return value of -1 indicates that there are no operational bridges on the ship
						// And hence it should be destroyed
		this._powerSystems();
		let best = -1; let neuralNet = false;
		for(let block of this._blocks) {
			if (block.key == "_b0" && block.operational) best = 0;
			if (block.key == "_b1" && block.operational) best = 1;
			if (block.key == "_b2" && block.operational) best = 2;
			if (block.key == "_b3" && block.operational) best = 3;
			if (block.key == "_b4" && block.operational) best = 4;
			if (block.key == "_b5" && block.operational) best = 5;
			if (block.key == "_b6" && block.operational) best = 6;
			if (block.key == "_b7" && block.operational) best = 7;
			if (block.key == "_b8" && block.operational) best = 8;
			if (block.key == "_nn" && block.operational) neuralNet = true;
		}
		if (best == -1) return -1;
		if (this._design.type === ShipDesignType.SpaceBase) {
			best++;
		}
		if (this._design.type === ShipDesignType.StarBase) {
			best = (best+1)*2;
		}
		if (neuralNet) {
			best *= 2;
		}
		return best;
	}
	hasABridge() {
		this._powerSystems();
		for(let block of this._blocks) {
			if (block.key == "_b0" && block.operational) return true;
			if (block.key == "_b1" && block.operational) return true;
			if (block.key == "_b2" && block.operational) return true;
			if (block.key == "_b3" && block.operational) return true;
			if (block.key == "_b4" && block.operational) return true;
			if (block.key == "_b5" && block.operational) return true;
			if (block.key == "_b6" && block.operational) return true;
			if (block.key == "_b7" && block.operational) return true;
			if (block.key == "_b8" && block.operational) return true;
		}
	}
	getColonyImprovement() {
		this._cacheBlocks();
		let colonyBlocks = 0;
		for(let block of this._blocks) {
			if (block.key == "_cb") colonyBlocks++;
		}
		return Math.floor(colonyBlocks/6);
	}
	hasWeapons(includeFighters) {
		this._powerSystems();
		for(let block of this._blocks) {
			if (block.key == "_wl" && block.operational) return true;
			if (block.key == "_wp" && block.operational) return true;
			if (block.key == "_wa" && block.operational) return true;
			if (block.key == "_wf" && block.operational) return true;
			if (block.key == "_wr" && block.operational) return true;
			if (block.key == "_wc" && block.operational) return true;
			if (block.key == "_ws" && block.operational) return true;
			if (block.key == "_wm" && block.operational) return true;
			if (block.key == "_wt" && block.operational) return true;
			if (block.key == "_wx" && block.operational) return true;
			if (block.key == "_wb" && block.operational) return true;
			if (block.key == "_df" && block.operational && includeFighters) return true;
		}
		return false;
	}
	getHits() {
		let hits = this._reflectedDamageThisPhase;
		this._powerSystems();
		for(let block of this._blocks) {
			if (block.key == "_wl" && block.operational && ROT.RNG.getUniform() < 0.166667) hits++;
			if (block.key == "_wp" && block.operational && ROT.RNG.getUniform() < 0.333334) hits++;
			if (block.key == "_wa" && block.operational && ROT.RNG.getUniform() < 0.5) hits++;
			if (block.key == "_wf" && block.operational) hits++;
			if (block.key == "_wr" && block.operational) hits += 10;
			if (block.key == "_wc" && block.operational) hits += 2;
			if (block.key == "_ws" && block.operational) hits += 20;
			if (block.key == "_wm" && block.operational) hits += 3;
			if (block.key == "_wt" && block.operational) hits += 30;
			if (block.key == "_wx" && block.operational) hits += 10;
			if (block.key == "_wb" && block.operational) hits += 1000000; // A million should be enough to destroy vulnerable targets
		}
		return hits;
	}
	getRepairCost() {
		this._cacheBlocks();
		return this._blocks.reduce((cost,block) => cost+block.repairCost(), 0);
	}
	getDestroyedBlocks() {
		this._cacheBlocks();
		return this._blocks.filter(block => block.repairCost()>0);
	}
	resolveHits(hits) {
		let report = "";
		// Save damage from reflecting shields
		if (this._reflectingShields) {
			this._reflectedDamageNextPhase += Math.min(hits,this._reflectingShields);
			this._reflectingShields -= Math.min(hits,this._reflectingShields);
		}
		// Take the damage on the shields if possible
		if (this._shields) {
			const points = Math.min(hits,this._shields);
			report += `${points} taken on the shields. `;
			this._shields -= points;
			hits -= points;
		}
		hitLoop: while(hits > 0 && this.hasABridge()) {
			// First take out any armour
			for(let block of this._blocks) {
				if (block.key == "_da" && block.operational) {
					block.addDamage();
					hits--;
					report += "Armour block destroyed. ";
					continue hitLoop;
				}
			}

			// OK then, now hit a random block (remembering that we need to account for large blocks)
			let choices = [];
			for(let block of this._blocks) {
				for(let i=0; i<block.size; i++) {
					if (!block.destroyed) {
						choices.push(block);
					}
				}
			}
			let choice = ROT.RNG.getItem(choices);
			choice.addDamage();
			if (choice.key == "_bb") {
				this._maxMarines--;
				this._marinesOnBoard = Math.min(this._marinesOnBoard,this._maxMarines);
			}
			hits--;
			this._powerNeedsCalculating = true;
			report += `${choice.name} hit. `;
		}
		return report;
	}
	repairAll() {
		this._cacheBlocks();
		for(let block of this._blocks) {
			block.repair();
		}
		this._maxMarines = this._design.getMarinesOnBoard(); // Barracks repaired (but the troops are not)
		this._powerNeedsCalculating = true;
	}
	getFighters() {
		this._powerSystems();
		let count = 0;
		for(let block of this._blocks) {
			if (block.key == "_df" && block.operational) count++;
		}
		return {count: count, spec: this._design.fighterSpec()};
	}
	getInvulnerable() {
		this._powerSystems();
		for(let block of this._blocks) {
			if (block.key == "_dv" && block.operational) return true;
		}
		return false;
	}
	raiseShieldsAtStartOfCombat() {
		this._powerSystems();
		this._shields = 0;
		this._reflectingShields = 0;
		this._reflectedDamageThisPhase = 0;
		this._reflectedDamageNextPhase = 0;
		for(let block of this._blocks) {
			if (block.key == "_ds" && block.operational) { this._shields += 10; }
			if (block.key == "_di" && block.operational) { this._shields += 20; }
			if (block.key == "_dr" && block.operational) { this._shields += 10; this._reflectingShields += 10; }
		}
	}
	raiseShieldsAtStartOfRound() {
		this._powerSystems();
		for(let block of this._blocks) {
			if (block.key == "_dg" && block.operational) { this._shields += 10; }
		}
	}
	nextCombatPhase() {
		this._reflectedDamageThisPhase = this._reflectedDamageNextPhase;
		this._reflectedDamageNextPhase = 0;
	}
	get size() {
		return this._design.totalSize();
	}
	get price() {
		return this._design.totalCost();
	}
	getScrapValue() {
		return Math.floor((this._design.totalCost() - this.getRepairCost())*0.4);
	}
	get marines() {
		return this._marinesOnBoard;
	}
	set marines(marines) {
		this._marinesOnBoard = marines;
	}
	getMaxMarines() {
		return this._maxMarines;
	}
	setTractor(points) {
		this._tractoredPoints = points;
	}
	_powerSystems() {
		if (this._powerNeedsCalculating) {
			this._cacheBlocks();
			let generated = 0; let generatedJump = 0;
			// Compute the power generated
			for(let block of this._blocks) {
				if (block.eps > 0) {
					block.setPowered(true);
					if (block.operational) generated += block.eps;
				}
				if (block.epsj > 0) {
					block.setPowered(true);
					if (block.operational) generatedJump += block.epsj;
				}
			}
			// And then power blocks (in block order, so the bridge first then drives followed by weapons and others)
			for(let block of this._blocks) {
				// Initially power the block (so that it will be operational (if not damaged) when we make the test)
				let need = (block.eps == "*50" ? 50 :block.eps); let needJump = block.epsj;
				if (need < 0) {
					block.setPowered(true);
					if (block.operational && (-need) <= generated) {
						generated += need;
					} else {
						block.setPowered(false);
					}
				}
				if (needJump < 0) {
					block.setPowered(true);
					if (block.operational && (-needJump) <= generatedJump) {
						generatedJump += needJump;
					} else {
						block.setPowered(false);
					}
				}
				if (need == 0 && needJump == 0) {
					block.setPowered(true); // Since it doesn't need power
				}
			}
			this._powerNeedsCalculating = false;
		}
	}
	_cacheBlocks() {
		// Need to deep clone the design's block's so that we have our own copy
		if (!this._blocks) {
			this._blocks = [];
			for(let block of this._design.getBlocks()) {
				this._blocks.push(new Block(block));
			}
		}
	}
	toJSON() {
		const data = {name: this._name,
			designId: this._design.id,
			loc: this._location,
			order: this._order,
			report: this._report,
			damage: null,
			retreatTo: this._retreatingTo,
			marines: this._maxMarines==0 ? null : this._marinesOnBoard+1 // To avoid the falsiness of 0
		};
		if (this._retreatingTo) {
			data.retreatTo = this._retreatingTo;
		}
		if (this._blocks && this._blocks.some(block => block.damage > 0)) {
			data.damage = "";
			for(let block of this._blocks) {
				if (block.damage === 0) {
					data.damage += " ";
				} else if (block.damage < 10) {
					data.damage += String.fromCharCode(48+block.damage);
				} else {
					data.damage += String.fromCharCode(54+block.damage); // @-sign=10, A-Z=11-36 etc up to r=50
				}
			}
		}
		return data;
	}
}
Ship.fromJSON = function(data) {
	const ship =  new Ship(data.name, ShipDesign.REGISTRY[data.designId],data.loc);
	ship._order = ShipOrder.fromJSON(data.order);
	ship._report = data.report;
	if (data.damage) {
		ship._cacheBlocks();
		for(let i=0; i<data.damage.length; i++) {
			if (data.damage[i] == " ") {
				// Noop
			} else if (data.damage[i] > "0" && data.damage[i] <= "9") {
				ship._blocks[i].damage = data.damage.charCodeAt(i)-48;
			} else {
				ship._blocks[i].damage = data.damage.charCodeAt(i)-54;
			}
		}
	}
	if (data.retreatTo) {
		ship._retreatingTo = data.retreatTo;
	}
	if (data.marines) {
		ship._marinesOnBoard = data.marines-1;
	}
	return ship;
};

//For testing via QUnit
if (typeof module !== "undefined" && module.exports) {
    exports.Ship = Ship;
}