/*!
 * © 2022 David Vines
 *
 * domainOfTheAncients is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * any later version.
 *
 * domainOfTheAncients is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with domainOfTheAncients. If not, see https://www.gnu.org/licenses/gpl-2.0.html.
 */
/* globals Game, Map, ROT, RepairShipOrder, ScrapShipOrder */
/* exported TurnProcessor */
var TurnProcessor = {
	processTurn: function() {
		const updates = [];
		const shipsToBeDestroyed = [];

		// Step 0. Clear all the colony reports
		for(let player of Game._players) {
			const selfMap = player.map;
			selfMap.forEachHex(hex => selfMap.setColonyReport(hex.id,null));
		}

		// Ship orders
		// Step 1. create a new map and place all ships in their new locations
		const workingMap = new Map(Game.State.baseMap);
		for(let player of Game._players) {
			for(let ship of player.ships) {
				ship.lastHex = ship.location;
				ship.retreatingTo = null;
				if (ship.order.jump) {
					ship.location = ship.order.jump;
				}
				workingMap.addObject(ship.location, {ship: ship, owner:player});
			}
		}

		// Step 2. Resolve what happening in each hex
		workingMap.forEachHex(hex => {
			// Step 2a. Work out what is in the hex
			ROT.RNG.shuffle(hex.objects);		// If there's no combat we need to ensure that we shuffle the ships to give the player with
												// the most colony ships the best chance to actually colonise the system first
			const bystyle = TurnProcessor.shipsByStyle(hex);	// Each element of bystyle (when present) is itself a sparse array showing how many ships of that size are present
			let troopsByStyle = {};
			if (hex.planet && hex.marines) troopsByStyle[hex.owner] = hex.marines;


			// Step 2b. Resolve combat
			if (Object.keys(bystyle).length > 1) {
				TurnProcessor.resolvePossibleCombat(hex,
					bystyle,
					(Object.keys(bystyle).some(style => style == Game.STYLENAMES[0])) ? updates : null,
					shipsToBeDestroyed); // Will remove destroyed ships from hex.objects
			}

			// Step 2c. Give the colony (if any) a report
			if (hex.owner && hex.level) {
				const report = TurnProcessor.reportOnOtherShips(hex,hex.owner);
				Game._stylenameToPlayerMapping[hex.owner].map.setColonyReport(hex.id,report);
				Game._stylenameToPlayerMapping[hex.owner].map.setLastScanned(hex.id,Game.State.turn,report);
			}

			// Step 2d. And then for each ship determine what it does
			for(let object of hex.objects) {
				if (object.ship && object.owner) {
					const ship = object.ship;
					const owner = object.owner;
					ship.report = "";

					// Scrap (if so ordered)
					if (ship.order instanceof ScrapShipOrder) {
						shipsToBeDestroyed.push(object);
						const actualRefund = ship.getScrapValue();
						ship.report = "Scrapped for "+actualRefund+" PPs. ";
						owner.carriedProduction += actualRefund;
					}

					// Repair orders
					if (ship.retreatingTo == null && ship.order instanceof RepairShipOrder) {
						if (ship.getRepairCost() == ship.order.value) {
							ship.repairAll();
							ship.report = "Fully repaired at a cost of ${ship.order.value}. ";
							owner.carriedProduction -= ship.order.value;
						} else if (ship.retreatingTo == null){
							ship.report = "Repair aborted as the ship was further damaged. ";
						}
					}

					// TODO: Planet Buster operation against planet

					// Colonisation (if ordered to do so)
					if (ship.order.colonise) {
						if (hex.planet && hex.owner !== owner.stylename && hex.level > 0) {
							ship.report += `Unable to colonise ${hex.id} as it has already been colonised by ${Game.FULLNAMES[hex.owner]}. `;
						} else if (hex.planet && hex.owner !== owner.stylename && hex.level == 0) {
							ship.report += `There is a star base belonging to ${Game.FULLNAMES[hex.owner]} here. Unable to colonise. `;
						} else if (hex.planet && hex.owner == owner.stylename && hex.level > 4) {
							ship.report += `Colonisation of ${hex.id} cannot be made as the colony is already at level ${hex.level}. `;
						} else if (hex.planet && hex.owner == owner.stylename && ship.getColonyImprovement() == 0) {
							ship.report += `Colonisation of ${hex.id} cannot be made as the colony ship does not have at least six undamaged colony blocks. `;
						} else if (hex.planet && hex.owner == owner.stylename && hex.level < 5) {
							ship.report += `Colonisation of ${hex.id} has been achieved adding ${ship.getColonyImprovement()} to the existing colony. `;
							Game.State.baseMap.colonise(ship.location,owner.stylename,ship.getColonyImprovement());
							Game.State.baseMap.addMarines(ship.location,ship.marines);
							hex.level += ship.getColonyImprovement();
							shipsToBeDestroyed.push(object);
						} else if (hex.planet) {
							ship.report += `Colonisation of ${hex.id} has been achieved. `;
							Game.State.baseMap.colonise(hex.id,owner.stylename,ship.getColonyImprovement());
							Game.State.baseMap.addMarines(ship.location,ship.marines);
							shipsToBeDestroyed.push(object);
						}
					}

					// Collect troops (if so ordered)
					if (ship.order.marinesEmbarking) {
						if (hex.planet) {
							if (hex.owner == owner.stylename) {
								if (hex.marines > 0) {
									if (hex.marines >= ship.getMaxMarines()-ship.marines) {
										troopsByStyle[hex.owner] += ship.marines-ship.getMaxMarines();
										hex.marines -= ship.marines-ship.getMaxMarines();
										Game.State.baseMap.addMarines(ship.location,ship.marines-ship.getMaxMarines()); // This removes the marines from the colony
										ship.marines = ship.getMaxMarines();
									} else {
										troopsByStyle[hex.owner] -= hex.marines;
										hex.marines = 0;
										ship.report += `Only ${hex.marines} marine compan${hex.marines == 1 ? "y" : "ies"} loaded as that was all that was available. `;
										Game.State.baseMap.setMarines(ship.location,0);
										ship.marines += hex.marines;
									}
								} else {
									ship.report += "No marines loaded as none were available. ";
								}
							} else {
								ship.report += "No marines loaded as we don't own this colony. '";
							}
						} else {
							ship.report += "No marines loaded as there is no planet here. ";
						}
					}

					// Disembark troops (if so ordered) {
					if (ship.order.marinesDisembarking && ship.marines > 0) {
						if (hex.planet) {
							troopsByStyle[owner.stylename] = troopsByStyle[owner.stylename] || 0;
							troopsByStyle[owner.stylename] += ship.marines;
							ship.marines = 0;
						} else {
							ship.report += "Marines have not disembarked as there is no planet here!";
						}
					}
				}
			}

			// Step 2e. Resolve ground combat (if any)
			if (Object.keys(troopsByStyle).length > 1) {
				const updatePlayer = Object.keys(troopsByStyle).some(style => style == Game.STYLENAMES[0]);
				let [owner,marines] = TurnProcessor.conductGroundCombat(hex.id,troopsByStyle,(updatePlayer ? updates : null));
				if (owner !== hex.owner) {
					if (owner == Game._players[0].stylename && hex.level > 1) {
						const report =`We have conquered ${hex.id} with  ${marines} marine compan${marines == 1 ? "y" : "ies"} remaining after the combat.`;
						updates.push(report);
						Game._stylenameToPlayerMapping[owner].map.setColonyReport(hex.id,report);
					} else if (owner == Game._players[0].stylename) {
						const report =`We have destroyed the level 1 colony at ${hex.id}.`;
						updates.push(report);
					}
					else if (hex.owner == Game._players[0].stylename) {
						const winner = Game.FULLNAMES[Game._stylenameToPlayerMapping[owner]];
						const report = `We have lost ${hex.id} to ${winner}`;
						updates.push(report);
					} else if (updatePlayer) {
						updates.push(`We did not conquer ${hex.id}.`);
					}
					if (hex.level > 1) {
						Game.State.baseMap.colonise(hex.id,owner,-1); // Changes owner and reduces the level by one
						hex.owner = owner; hex.level--; // And update the copy we're using in this turn processor
						Game.State.baseMap.setMarines(hex.id,marines);
						hex.marines = marines;
					} else {
						delete hex.owner; delete hex.level;
						Game.State.baseMap.removeColony(hex.id);
						delete hex.owner; delete hex.level; delete hex.marines; // And update the copy we're using this turn processor
					}
				} else {
					Game.State.baseMap.setMarines(hex.id,marines);
					hex.marines = marines;
				}
			} else if (Object.keys(troopsByStyle).length == 1) {
				Game.State.baseMap.setMarines(hex.id,troopsByStyle[hex.owner]); // Put the marines (plus any additions) back on the planet
				hex.marines = troopsByStyle[hex.owner];
			}

			// Step 2f. Finally generate ship reports (after colonisation and ground combat)
			for(let object of hex.objects) {
				if (object.ship && object.owner) {
					const ship = object.ship;
					const owner = object.owner;

					// Scanning of the hex
					if (hex.planet && hex.owner) {
						ship.report += `${ship.location} is a class ${hex.level} colony belonging to ${Game.FULLNAMES[hex.owner]}. There ${hex.marines == 1 ? "is" : "are"} ${hex.marines} marine compan${hex.marines == 1? "y" : "ies"} here. `;
					} else if (hex.planet) {
						ship.report += `${ship.location} is an uncolonised world. `;
					}

					ship.report += TurnProcessor.reportOnOtherShips(hex,owner.stylename);

					// Update the hex report before the ship is destroyed (if it is going to be destroyed)
					owner.map.setLastScanned(hex.id,Game.State.turn,ship.report); // If a player has multiple ships we will set this multiple times (to the same value)
				}
			}
		});

		// Step 3. Remove those ship no longer in existence
		for(let object of shipsToBeDestroyed) {
			object.owner.destroyShip(object.ship);
			if (object.owner == (Game._players[0])) {
				updates.push(object.ship.report); // Since there's no other way the player will see the report!
			}
		}

		// Step 4. Apply non-ship orders
		for(let i=0; i<Game._players.length; i++) {
			const stylename = Game._players[i].stylename;
			for(let order of Game._orders[stylename]) {
				let update = order.apply(Game._players[i],Game.State.baseMap,Game.State.turn,i===0);
				if (update) {
					// There's an update and since it's for the human we want to add it to the list of moves
					updates.push(update);
				}
			}
		}
		return updates;
	},
	shipsByStyle(hex) {
		let bystyle = {}; // (re)build the mapping from stylename to a sparse array of ship counts (of a given ship size) present in the hex
		for(let object of hex.objects) {
			if (object.ship && object.owner) {
				const name = object.owner.stylename;
				bystyle[name] = bystyle[name] ? bystyle[name] : [];
				bystyle[name][0] = bystyle[name][0] ? bystyle[name][0] : 1; // Use the 0 entry to count all ships
				bystyle[name][object.ship.size] = bystyle[name][object.ship.size] ? bystyle[name][object.ship.size]++ : 1;
			}
		}
		return bystyle;
	},
	reportOnOtherShips: function(hex,owner) {
		// TODO: Worry about the after effects of combat - we may need to pass around/recalculate after destroying ships
		const bystyle = TurnProcessor.shipsByStyle(hex);
		delete bystyle[owner];
		let report = "";
		for(let otherowner of Object.keys(bystyle)) {
			for(let size=1; size<bystyle[otherowner].length; size++) {
				if (bystyle[otherowner][size]) {
					report += "There " +
						(bystyle[otherowner][size]==1 ? "is a" : "are "+bystyle[otherowner][size]) +
						" size "+size+" ship" +
						(bystyle[otherowner][size]==1 ? "" : "s") +
						" belonging to " +
						Game.FULLNAMES[otherowner] +
						" here. ";
				}
			}
		}
		return report;
	},
	resolvePossibleCombat: function(hex,bystyle,updates,shipsToBeDestroyed) {
		const styles = Object.keys(bystyle);
		let from = "";
		for(let i=0; i<styles.length; i++) {
			if (i>0 && i<styles.length-1) {
				from += ", ";
			} else if (i == styles.length-1) {
				from += " and ";
			}
			from += Game.FULLNAMES[styles[i]];
		}
		let list = "";
		for(let i=0; i<hex.objects.length; i++) {
			if (i>0 && i<hex.objects.length-1) {
				list += ", ";
			} else if (i == hex.objects.length-1) {
				list += " and ";
			}
			list += hex.objects[i].ship.name;
		}

		// Does any ship have any weapons?
		if (hex.objects.some(object => object.ship.hasWeapons(true))) {
			if (updates) {updates.push(`Combat!!!: In ${hex.id} there are ships from ${from}: ${list}`); }
			TurnProcessor.resolveCombat(hex,updates,shipsToBeDestroyed);
		}
	},
	whosPresent: function(hex) {
			const bystyle = {}; // each element (when present) has a sparse array showing how many ships of that size are present
			for(let object of hex.objects) {
				if (object.ship && object.owner) {
					const name = object.owner.stylename;
					bystyle[name] = bystyle[name] ? bystyle[name] : true;
				}
			}
		return Object.keys(bystyle);
	},
	// TODO: Take account of any installed predictors
	// TODO: Update ship report with the result of the fight
	resolveCombat(hex,updates,shipsToBeDestroyed) {
		let maxPhaseNumber = 1;
		for(let object of hex.objects) {
			maxPhaseNumber = Math.max(maxPhaseNumber, object.ship.getCombatSpeed());
			// Also do the start of combat preparations
			object.ship.raiseShieldsAtStartOfCombat();
		}
		if (updates) { updates.push(`This combat will have ${maxPhaseNumber} phase(s) per combat round`); }

		// TODO: launch fighters
		// TODO: Sort out invulnerable targets

		let round = 0;
		while((TurnProcessor.whosPresent(hex).length > 1) && (hex.objects.some(object => object.ship.hasWeapons(false)))) {
			round++;
			for(let phase=maxPhaseNumber; phase>0; phase--) {
				let hitTargets = [];
				let hit = [];
				for(let object of hex.objects) {
					if (!object.ship) continue; // It's not a ship
					const ship = object.ship;
					if (!object.ship.hasABridge()) continue; // It's lost all of its bridges
					if (ship.getCombatSpeed() < phase) continue; // Too slow, this ship cannot act
					ship.nextCombatPhase(); // Let the ship use its reflecting shields
					let targets = [];
					for(let other of hex.objects) {
						if (other.ship && other.owner !== object.owner) { targets.push(other.ship); }
					}
					if (targets.length > 0) {
						const target = ROT.RNG.getItem(targets);
						const hits = ship.getHits();
						if (hits > 0) {
							if (updates) { updates.push(`Round ${round} Phase ${phase}: ${ship.name} scores ${hits} hits on ${target.name}`); }
							hitTargets.push(target); hit.push(hits);
						}
					}
				}
				for(let i=0; i<hitTargets.length; i++) {
					let resolution = hitTargets[i].resolveHits(hit[i]);
					if (updates) { updates.push(`Round ${round} Phase ${phase}: ${hitTargets[i].name} ${resolution}`); }
				}
				// Resolve destroyed ships
				let newlist = [];
				for(let object of hex.objects) {
					if (object.ship && !object.ship.hasABridge()) {
						shipsToBeDestroyed.push(object);
						if (updates) { updates.push(`Round ${round}: ${object.ship.name} has been destroyed`); }
					} else{
						// The ship is still in the fight
						newlist.push(object);
					}
				}
				hex.objects = newlist;
			}

			// Resolve retreats if there's still more than one side
			if ((TurnProcessor.whosPresent(hex).length > 1) && (hex.objects.some(object => object.ship.hasWeapons(false)))) {
				let newlist = [];
				for(let object of hex.objects) {
					if (object.ship && !object.ship.hasWeapons(false) && object.ship.getJumpRange() > 0) {
						// Retreat the ship
						object.ship.retreatingTo = object.ship.lastHex;
						if (updates) { updates.push(`Round ${round}: ${object.ship.name} is retreating to ${object.ship.lastHex}`); }
					} else {
						// The ship is still in the fight
						newlist.push(object);
					}
				}
				hex.objects = newlist;
			}
		}
	},
	conductGroundCombat(location,troopsByStyle,updates) {
		if (updates) { updates.push(`Ground Combat at ${location}: ${TurnProcessor.listTroops(troopsByStyle)}`); }
		// Multiply up the companies by the planetary defence tech
		for(let owner of Object.keys(troopsByStyle)) {
			troopsByStyle[owner] *= Game._stylenameToPlayerMapping[owner].troopValue;
		}

		let round = 0;
		while(Object.keys(troopsByStyle).length > 1) {
			round++;
			let newTroopsByStyle = Object.assign({},troopsByStyle);
			const keys = Object.keys(newTroopsByStyle);
			for(let owner of keys) {
				for(let opponent of keys) {
					if (owner !== opponent) {
						newTroopsByStyle[opponent] = Math.max(0, newTroopsByStyle[opponent] - Math.floor(troopsByStyle[owner] * ROT.RNG.getPercentage()/100));
					}
				}
			}
			for(let owner of keys) {
				if (newTroopsByStyle[owner] == 0) {
					delete newTroopsByStyle[owner];
				}
			}
			troopsByStyle = newTroopsByStyle;
		}
		const winner = Object.keys(troopsByStyle)[0];
		return [winner, Math.floor(troopsByStyle[winner]/Game._stylenameToPlayerMapping[winner].troopValue)];
	},
	listTroops(troopsByStyle) {
		const keys = Object.keys(troopsByStyle);
		let list = "There ";
		for(let i=0; i<keys.length; i++) {
			const owner = keys[i];
			if (i == keys.length-1) {
				list +=" and ";
			} else if (i >0) {
				list += ", ";
			} else if (troopsByStyle[owner] == 1) {
				list += "is ";
			} else {
				list += "are ";
			}
			if (troopsByStyle[owner] == 1) {
				list += "a marine company from ";
			} else {
				list += troopsByStyle[owner]+" marine companies from ";
			}
			list += Game.FULLNAMES[owner];
		}
		return list;
	}
};